"""Control components."""

from typing import Callable, Optional, List, Union, Sequence

from bowtie._component import Component, jdumps, jsbool


# pylint: disable=too-few-public-methods
class _Controller(Component):
    """Abstract class all control components inherit.

    Used to test if a an object is a controller.
    """

    # pylint: disable=abstract-method
    @property
    def _instantiate(self) -> str:
        tagwrap = '{component}' + self._tagbase
        return self._insert(tagwrap, self._comp)


class Button(_Controller):
    """An Ant design button."""

    _TEMPLATE = 'button.jsx'
    _COMPONENT = 'SimpleButton'
    _PACKAGE = None
    _ATTRS = "label={{'{label}'}}"

    def __init__(self, label: str = '') -> None:
        """Create a button.

        Parameters
        ----------
        label : str, optional
            Label on the button.

        """
        super().__init__()
        self._comp = self._tag.format(
            label=label
        )

    def on_click(self) -> None:
        """Emit an event when the button is clicked.

        There is no getter associated with this event.

        | **Payload:** ``None``.

        Returns
        -------
        str
            Name of click event.

        """


class Upload(_Controller):
    """Draggable file upload widget."""

    _TEMPLATE = 'upload.jsx'
    _COMPONENT = 'AntUpload'
    _PACKAGE = None
    _ATTRS = "multiple={{{multiple}}}"

    def __init__(self, multiple=True):
        """Create the widget.

        Note: the handler parameter may be changed in the future.

        Parameters
        ----------
        multiple : bool, optional
            If true, you can upload multiple files at once. Even with this set to true,
            the handler will get called once per file uploaded.

        """
        super().__init__()
        self._comp = self._tag.format(
            multiple=jsbool(multiple)
        )

    def on_upload(self):
        """Emit an event when the selection changes.

        There is no getter associated with this event.

        | **Payload:** ``tuple`` with a str (name) and BytesIO (stream).

        The user is responsible for storing the object in this function
        if they want it for later use. To indicate an error, return True,
        otherwise a return value of None or False indicate success.

        """


class Dropdown(_Controller):
    """Dropdown based on react-select."""

    _TEMPLATE = 'dropdown.jsx'
    _COMPONENT = 'Dropdown'
    _PACKAGE = 'react-select@1.3.0'
    _ATTRS = ('initOptions={{{options}}} '
              'multi={{{multi}}} '
              'default={{{default}}}')

    def __init__(self, labels: Optional[List[str]] = None,
                 values: Optional[List[Union[str, int]]] = None, multi: bool = False,
                 default: Optional[Union[str, int]] = None) -> None:
        """Create a drop down.

        Parameters
        ----------
        labels : array-like, optional
            List of strings which will be visible to the user.
        values : array-like, optional
            List of values associated with the labels that are hidden from the user.
        multi : bool, optional
            If multiple selections are allowed.
        default : str or int, optional
            The default selected value.

        """
        super().__init__()

        if values is not None and labels is not None:
            options = [{'value': value, 'label': str(label)}
                       for value, label in zip(values, labels)]
        else:
            options = []

        self._comp = self._tag.format(
            options=jdumps(options),
            multi=jsbool(multi),
            default=jdumps(default),
        )

    def on_change(self) -> Callable:
        """Emit an event when the selection changes.

        | **Payload:** ``dict`` with keys "value" and "label".

        """
        return self.get

    # pylint: disable=no-self-use
    def do_options(self, labels, values):
        """Replace the drop down fields.

        Parameters
        ----------
        labels : array-like
            List of strings which will be visible to the user.
        values : array-like
            List of values associated with the labels that are hidden from the user.

        Returns
        -------
        None

        """
        return [dict(label=l, value=v) for l, v in zip(labels, values)]

    # pylint: disable=no-self-use
    def do_choose(self, values):
        """Replace the drop down fields.

        Parameters
        ----------
        values : list or str or int
            Value(s) of drop down item(s) to be selected.

        Returns
        -------
        None

        """
        return values

    def get(self, data):
        """Return selected value(s)."""
        return data


class Switch(_Controller):
    """Toggle switch."""

    _TEMPLATE = 'switch.jsx'
    _COMPONENT = 'Toggle'
    _PACKAGE = None
    _ATTRS = 'defaultChecked={{{defaultChecked}}}'

    def __init__(self, initial: bool = False) -> None:
        """Create a toggle switch.

        Parameters
        ----------
        initial : bool, optional
            Starting state of the switch.

        """
        super().__init__()
        self._comp = self._tag.format(
            defaultChecked=jsbool(initial)
        )

    def on_switch(self):
        """Emit an event when the switch is toggled.

        | **Payload:** ``bool`` status of the switch.

        Returns
        -------
        str
            Name of event.

        """
        return self.get

    # pylint: disable=no-self-use
    def get(self, data):
        """
        Get the state of the switch.

        Returns
        -------
        bool
            True if the switch is enabled.

        """
        return data


class _DatePickers(_Controller):
    """Specific Date Pickers inherit this class."""

    _TEMPLATE = 'date.jsx'
    _COMPONENT = 'PickDates'
    _PACKAGE = None
    _ATTRS = ('date={{{date_type}}} '
              'month={{{month_type}}} '
              'range={{{range_type}}}')

    def __init__(self, date_type: bool = False, month_type: bool = False,
                 range_type: bool = False) -> None:
        super().__init__()
        self._comp = self._tag.format(
            date_type=jsbool(date_type),
            month_type=jsbool(month_type),
            range_type=jsbool(range_type)
        )


class DatePicker(_DatePickers):
    """A Date Picker.

    Let's you choose an individual day.
    """

    def __init__(self) -> None:
        """Create a date picker."""
        super().__init__(date_type=True)

    def on_change(self):
        """Emit an event when a date is selected.

        | **Payload:** ``str`` of the form ``"yyyy-mm-dd"``.

        Returns
        -------
        str
            Name of event.

        """
        return self.get

    # pylint: disable=no-self-use
    def get(self, data):
        """
        Get the currently selected date.

        Returns
        -------
        str
            Date in the format "YYYY-MM-DD"

        """
        return data


class MonthPicker(_DatePickers):
    """A Month Picker.

    Let's you choose a month and year.
    """

    def __init__(self) -> None:
        """Create month picker."""
        super().__init__(month_type=True)

    def on_change(self):
        """Emit an event when a month is selected.

        | **Payload:** ``str`` of the form ``"yyyy-mm"``.

        Returns
        -------
        str
            Name of event.

        """
        return self.get

    # pylint: disable=no-self-use
    def get(self, data):
        """
        Get the currently selected month.

        Returns
        -------
        str
            Month in the format "YYYY-MM"

        """
        return data


class RangePicker(_DatePickers):
    """A Date Range Picker.

    Choose two dates to use as a range.
    """

    def __init__(self) -> None:
        """Create a date range picker."""
        super().__init__(range_type=True)

    def on_change(self):
        """Emit an event when a range is selected.

        | **Payload:** ``list`` of two dates ``["yyyy-mm-dd", "yyyy-mm-dd"]``.

        Returns
        -------
        str
            Name of event.

        """
        return self.get

    # pylint: disable=no-self-use
    def get(self, data):
        """
        Get the currently selected date range.

        Returns
        -------
        list
            A list of two strings ``["yyyy-mm-dd", "yyyy-mm-dd"]``.

        """
        return data


class Number(_Controller):
    """A number input widget with increment and decrement buttons."""

    _TEMPLATE = 'number.jsx'
    _COMPONENT = 'AntNumber'
    _PACKAGE = None
    _ATTRS = ('start={{{start}}} '
              'min={{{minimum}}} '
              'max={{{maximum}}} '
              'step={{{step}}} '
              "size={{'{size}'}}")

    def __init__(self, start: int = 0, minimum: float = -1e100, maximum: float = 1e100,
                 step: int = 1, size: str = 'default') -> None:
        """Create a number input.

        Parameters
        ----------
        start : number, optional
            Starting number
        minimum : number, optional
            Lower bound
        maximum : number, optional
            Upper bound
        size : 'default', 'large', 'small', optional
            Size of the text box.

        References
        ----------
        https://ant.design/components/input/

        """
        super().__init__()
        self._comp = self._tag.format(
            start=start,
            minimum=minimum,
            maximum=maximum,
            step=step,
            size=size
        )

    def on_change(self):
        """Emit an event when the number is changed.

        | **Payload:** ``number``

        Returns
        -------
        str
            Name of event.

        """
        return self.get

    # pylint: disable=no-self-use
    def get(self, data):
        """
        Get the current number.

        Returns
        -------
        number

        """
        return data


class Textbox(_Controller):
    """A single line text box."""

    _TEMPLATE = 'textbox.jsx'
    _COMPONENT = 'Textbox'
    _PACKAGE = None
    _ATTRS = ("placeholder={{'{placeholder}'}} "
              "size={{'{size}'}} "
              "type={{'{area}'}} "
              'autosize={{{autosize}}} '
              'disabled={{{disabled}}}')

    def __init__(self, placeholder: str = 'Enter text', size: str = 'default', area: bool = False,
                 autosize: bool = False, disabled: bool = False) -> None:
        """Create a text box.

        Parameters
        ----------
        placeholder : str, optional
            Initial text that appears.
        size : 'default', 'large', 'small', optional
            Size of the text box.
        area : bool, optional
            Create a text area if True else create a single line input.
        autosize : bool, optional
            Automatically size the widget based on the content.
        disabled : bool, optional
            Disable input to the widget.

        References
        ----------
        https://ant.design/components/input/

        """
        super().__init__()

        self._comp = self._tag.format(
            area='textarea' if area else 'text',
            autosize=jsbool(autosize),
            disabled=jsbool(disabled),
            placeholder=placeholder,
            size=size
        )

    # pylint: disable=no-self-use
    def do_text(self, text):
        """Set the text of the text box.

        Parameters
        ----------
        text : str
            String of the text box.

        """
        return text

    def on_enter(self):
        """Emit an event when enter is pressed in the text box.

        | **Payload:** ``str``

        Returns
        -------
        str
            Name of event.

        """
        return self.get

    def on_change(self) -> Callable:
        """Emit an event when the text is changed.

        | **Payload:** ``str``

        Returns
        -------
        str
            Name of event.

        """
        return self.get

    # pylint: disable=no-self-use
    def get(self, data):
        """
        Get the current text.

        Returns
        -------
        str

        """
        return data


class Slider(_Controller):
    """Ant Design slider."""

    _TEMPLATE = 'slider.jsx'
    _COMPONENT = 'AntSlider'
    _PACKAGE = None
    _ATTRS = ('range={{{range}}} '
              'min={{{minimum}}} '
              'max={{{maximum}}} '
              'step={{{step}}} '
              'start={{{start}}} '
              'marks={{{marks}}} '
              'vertical={{{vertical}}}')

    def __init__(self, start: Optional[Union[float, Sequence[float]]] = None, ranged: bool = False,
                 minimum: float = 0, maximum: float = 100, step: float = 1,
                 vertical: bool = False) -> None:
        """Create a slider.

        Parameters
        ----------
        start : number or list with two values, optional
            Determines the starting value.
            If a list of two values are given it will be a range slider.
        ranged : bool, optional
            If this is a range slider.
        minimum : number, optional
            Minimum value of the slider.
        maximum : number, optional
            Maximum value of the slider.
        step : number, optional
            Step size.
        vertical : bool, optional
            If True, the slider will be vertical

        References
        ----------
        https://ant.design/components/slider/

        """
        super().__init__()

        if start is None:
            if ranged:
                start = [minimum, maximum]
            else:
                start = minimum
        elif isinstance(start, Sequence):
            if len(start) > 2:
                raise ValueError('start cannot be more than 2 numbers')
            start = list(start)
            ranged = True

        self._comp = self._tag.format(
            range=jsbool(ranged),
            minimum=minimum,
            maximum=maximum,
            start=start,
            step=step,
            marks={minimum: str(minimum), maximum: str(maximum)},
            vertical=jsbool(vertical)
        )

    # pylint: disable=no-self-use
    def do_max(self, value):
        """Replace the max value of the slider.

        Parameters
        ----------
        value : int
            Maximum value of the slider.

        """
        return value

    def do_min(self, value):
        """Replace the min value of the slider.

        Parameters
        ----------
        value : int
            Minimum value of the slider.

        """
        return value

    def do_value(self, value):
        """Set the value of the slider.

        Parameters
        ----------
        value : int
            Value of the slider.

        """
        return value

    def do_inc(self, value=1):
        """Increment value of slider by given amount.

        Parameters
        ----------
        value : int
            Number to change value of slider by.

        """
        return value

    def do_min_max_value(self, minimum, maximum, value):
        """Set the minimum, maximum, and value of slider simultaneously.

        Parameters
        ----------
        minimum : int
            Minimum value of the slider.
        maximum : int
            Maximum value of the slider.
        value : int
            Value of the slider.

        """
        return minimum, maximum, value

    def on_change(self) -> Callable:
        """Emit an event when the slider's value changes.

        | **Payload:** ``number`` or ``list`` of values.

        Returns
        -------
        str
            Name of event.

        """
        return self.get

    def on_after_change(self):
        """Emit an event when the slider control is released.

        | **Payload:** ``number`` or ``list`` of values.

        Returns
        -------
        str
            Name of event.

        """
        return self.get

    def get(self, data):
        """
        Get the currently selected value(s).

        Returns
        -------
        list or number
            List if it's a range slider and gives two values.

        """
        return data


class Nouislider(_Controller):
    """A lightweight JavaScript range slider library."""

    _TEMPLATE = 'nouislider.jsx'
    _COMPONENT = 'Nouislider'
    _PACKAGE = 'nouislider@11.1.0'
    _ATTRS = ('range={{{{min: {min}, max: {max}}}}} '
              'socket={{socket}} '
              'start={{{start}}} '
              'tooltips={{{tooltips}}}')

    def __init__(self, start: Union[int, Sequence[int]] = 0, minimum: int = 0,
                 maximum: int = 100, tooltips: bool = True) -> None:
        """Create a slider.

        Parameters
        ----------
        start : number or list with two values, optional
            Determines the starting value.
            If a list of two values are given it will be a range slider.
        minimum : number, optional
            Minimum value of the slider.
        maximum : number, optional
            Maximum value of the slider.
        tooltips : bool, optional
            Show a popup text box.

        References
        ----------
        https://refreshless.com/nouislider/events-callbacks/

        """
        super().__init__()

        if not isinstance(start, Sequence):
            nstart = [start]
        else:
            nstart = list(start)
        self._comp = self._tag.format(
            min=minimum,
            max=maximum,
            start=nstart,
            tooltips=jsbool(tooltips)
        )

    def on_update(self):
        """Emit an event when the slider is moved.

        https://refreshless.com/nouislider/events-callbacks/

        | **Payload:** ``list`` of values.

        Returns
        -------
        str
            Name of event.

        """
        return self.get

    def on_slide(self):
        """Emit an event when the slider is moved.

        https://refreshless.com/nouislider/events-callbacks/

        | **Payload:** ``list`` of values.

        Returns
        -------
        str
            Name of event.

        """
        return self.get

    def on_set(self):
        """Emit an event when the slider is moved.

        https://refreshless.com/nouislider/events-callbacks/

        | **Payload:** ``list`` of values.

        Returns
        -------
        str
            Name of event.

        """
        return self.get

    def on_change(self) -> Callable:
        """Emit an event when the slider is moved.

        https://refreshless.com/nouislider/events-callbacks/

        | **Payload:** ``list`` of values.

        Returns
        -------
        str
            Name of event.

        """
        return self.get

    def on_start(self):
        """Emit an event when the slider is moved.

        https://refreshless.com/nouislider/events-callbacks/

        | **Payload:** ``list`` of values.

        Returns
        -------
        str
            Name of event.

        """
        return self.get

    def on_end(self):
        """Emit an event when the slider is moved.

        https://refreshless.com/nouislider/events-callbacks/

        | **Payload:** ``list`` of values.

        Returns
        -------
        str
            Name of event.

        """
        return self.get

    # pylint: disable=no-self-use
    def get(self, data):
        """
        Get the currently selected value(s).

        Returns
        -------
        list or number
            List if it's a range slider and gives two values.

        """
        return data



"""Messages provide a temporary message that will disappear after a few seconds.

Reference
^^^^^^^^^
https://ant.design/components/message/
"""

import flask
from flask_socketio import emit
import eventlet

from bowtie._component import pack
from bowtie._utils import func_name


def _message(status, content):
    """Send message interface.

    Parameters
    ----------
    status : str
        The type of message
    content : str

    """
    event = f'message.{status}'
    if flask.has_request_context():
        emit(event, dict(data=pack(content)))
    else:
        sio = flask.current_app.extensions['socketio']
        sio.emit(event, dict(data=pack(content)))
    eventlet.sleep()


def success(content):
    """Success message.

    Parameters
    ----------
    content : str
        Message to show user.

    """
    _message(func_name(), content)


def error(content):
    """Error message.

    Parameters
    ----------
    content : str
        Message to show user.

    """
    _message(func_name(), content)


def info(content):
    """Info message.

    Parameters
    ----------
    content : str
        Message to show user.

    """
    _message(func_name(), content)


def warning(content):
    """Warning message.

    Parameters
    ----------
    content : str
        Message to show user.

    """
    _message(func_name(), content)


def loading(content):
    """Load message.

    Parameters
    ----------
    content : str
        Message to show user.

    """
    _message(func_name(), content)



"""Utility functions for testing only."""

from contextlib import contextmanager
from multiprocessing import Process
import time

from bowtie import View
from bowtie._component import Component


def reset_uuid():
    """Reset the uuid counter for components."""
    # pylint: disable=protected-access
    Component._NEXT_UUID = 0
    View._NEXT_UUID = 0


@contextmanager
def server_check(app):
    """Context manager for testing Bowtie apps and verifying no errors happened."""
    process = Process(target=app._serve)  # pylint: disable=protected-access
    process.start()
    time.sleep(5)
    yield process
    process.terminate()



"""Visual components."""

from typing import Dict, Optional, List, Union, Tuple

from bowtie._component import Component, jdumps, jsbool
from bowtie._progress import Progress


# pylint: disable=too-few-public-methods
class _Visual(Component):
    """Abstract class all visual components inherit.

    Used to test if a an object is a visual component.
    """

    # pylint: disable=abstract-method
    def __init__(self) -> None:
        self.progress = Progress()
        super().__init__()

    @property
    def _instantiate(self) -> str:
        # pylint: disable=protected-access
        begin, end = self.progress._tags
        tagwrap = begin + '{component}' + self._tagbase + end
        return self._insert(tagwrap, self._comp)


class Table(_Visual):
    """Ant Design table with filtering and sorting."""

    _TEMPLATE = 'table.jsx'
    _COMPONENT = 'AntTable'
    _PACKAGE = None
    _ATTRS = ('columns={{{columns}}} '
              'resultsPerPage={{{results_per_page}}}')

    def __init__(self, data=None, columns: Optional[List[Union[int, str]]] = None,
                 results_per_page: int = 10) -> None:
        """Create a table and optionally initialize the data.

        Parameters
        ----------
        data : pd.DataFrame, optional
        columns : list, optional
            List of column names to display.
        results_per_page : int, optional
            Number of rows on each pagination of the table.

        """
        super().__init__()
        self.data = []  # type: List[Dict]
        self.columns = []  # type: List[Dict]
        if data:
            self.data, self.columns = self._make_data(data)
        elif columns:
            self.columns = self._make_columns(columns)

        self.results_per_page = results_per_page

        self._comp = self._tag.format(
            columns=self.columns,
            results_per_page=self.results_per_page
        )

    @staticmethod
    def _make_columns(columns: List[Union[int, str]]) -> List[Dict]:
        """Transform list of columns into AntTable format."""
        return [dict(title=str(c),
                     dataIndex=str(c),
                     key=str(c))
                for c in columns]

    @staticmethod
    def _make_data(data) -> Tuple[List[Dict], List[Dict]]:
        """Transform table data into JSON."""
        jsdata = []
        for idx, row in data.iterrows():
            row.index = row.index.astype(str)
            rdict = row.to_dict()
            rdict.update(dict(key=str(idx)))
            jsdata.append(rdict)

        return jsdata, Table._make_columns(data.columns)

    # pylint: disable=no-self-use
    def do_data(self, data):
        """Replace the columns and data of the table.

        Parameters
        ----------
        data : pandas.DataFrame

        Returns
        -------
        None

        """
        return self._make_data(data)

    def do_columns(self, columns):
        """Update the columns of the table.

        Parameters
        ----------
        columns : array-like
            List of strings.

        Returns
        -------
        None

        """
        return self._make_columns(columns)


class SmartGrid(_Visual):
    """Griddle table with filtering and sorting."""

    _TEMPLATE = 'griddle.jsx'
    _COMPONENT = 'SmartGrid'
    _PACKAGE = 'griddle-react@1.11.2'
    _ATTRS = None

    def __init__(self) -> None:
        """Create the table, optionally set the columns."""
        super().__init__()
        self._comp = self._tag

    # pylint: disable=no-self-use
    def do_update(self, data):
        """Update the data of the table.

        Parameters
        ----------
        data : pandas.DataFrame

        Returns
        -------
        None

        """
        return data.to_dict(orient='records')

    def get(self, data):
        """
        Get the table data.

        Returns
        -------
        list
            Each entry in the list is a dict of labels and values for a row.

        """
        return data


class SVG(_Visual):
    """SVG image.

    Mainly for matplotlib plots.
    """

    _TEMPLATE = 'svg.jsx'
    _COMPONENT = 'SVG'
    _PACKAGE = None
    _ATTRS = 'preserveAspectRatio={{{preserve_aspect_ratio}}}'

    def __init__(self, preserve_aspect_ratio: bool = False) -> None:
        """Create SVG component.

        Parameters
        ----------
        preserve_aspect_ratio : bool, optional
            If ``True`` it preserves the aspect ratio otherwise
            it will stretch to fill up the space available.

        """
        super().__init__()
        self.preserve_aspect_ratio = preserve_aspect_ratio
        self._comp = self._tag.format(
            preserve_aspect_ratio=jsbool(self.preserve_aspect_ratio)
        )

    # pylint: disable=no-self-use
    def do_image(self, image):
        """Replace the image.

        Parameters
        ----------
        image : str
            Generated by ``savefig`` from matplotlib with ``format=svg``.

        Returns
        -------
        None

        Examples
        --------
        This shows how to update an SVG widget with Matplotlib.

        >>> from io import StringIO
        >>> import matplotlib
        >>> matplotlib.use('Agg')
        >>> import matplotlib.pyplot as plt
        >>> image = SVG()
        >>>
        >>> def callback(x):
        ...     sio = StringIO()
        ...     plt.plot(range(5))
        ...     plt.savefig(sio, format='svg')
        ...     sio.seek(0)
        ...     s = sio.read()
        ...     idx = s.find('<svg')
        ...     s = s[idx:]
        ...     image.do_image(s)

        """
        return image


class Plotly(_Visual):
    """Plotly component.

    Useful for many kinds of plots.
    """

    _TEMPLATE = 'plotly.jsx'
    _COMPONENT = 'PlotlyPlot'
    _PACKAGE = 'plotly.js@1.40.1'
    _ATTRS = 'initState={{{init}}}'

    def __init__(self, init: Optional[Dict] = None) -> None:
        """Create a Plotly component.

        Parameters
        ----------
        init : dict, optional
            Initial Plotly data to plot.

        """
        super().__init__()
        if init is None:
            init = dict(data=[], layout={'autosize': False})
        self.init = init

        self._comp = self._tag.format(
            init=jdumps(self.init)
        )

    # Events

    def on_click(self):
        """React to Plotly click event.

        | **Payload:** TODO.

        Returns
        -------
        str
            Name of event.

        """
        return self.get_click

    def on_beforehover(self):
        """Emit an event before hovering over a point.

        | **Payload:** TODO.

        Returns
        -------
        str
            Name of event.

        """
        return self.get_hover

    def on_hover(self):
        """Emit an event after hovering over a point.

        | **Payload:** TODO.

        Returns
        -------
        str
            Name of event.

        """
        return self.get_hover

    def on_unhover(self):
        """Emit an event when hover is removed.

        | **Payload:** TODO.

        Returns
        -------
        str
            Name of event.

        """
        return self.get_hover

    def on_select(self):
        """Emit an event when points are selected with a tool.

        | **Payload:** TODO.

        Returns
        -------
        str
            Name of event.

        """
        return self.get_select

    def on_relayout(self):
        """Emit an event when the chart axes change.

        | **Payload:** TODO.

        Returns
        -------
        str
            Name of event.

        """
        return self.get_layout

    # Commands

    # pylint: disable=no-self-use
    def do_all(self, plot):
        """Replace the entire plot.

        Parameters
        ----------
        plot : dict
            Dict that can be plotted with Plotly.
            It should have this structure: ``{data: [], layout: {}}``.

        Returns
        -------
        None

        """
        return plot

    def do_data(self, data):
        """Replace the data portion of the plot.

        Parameters
        ----------
        data : list of traces
            List of data to replace the old data.

        Returns
        -------
        None

        """
        return data

    def do_layout(self, layout):
        """Update the layout.

        Parameters
        ----------
        layout : dict
            Contains layout information.

        Returns
        -------
        None

        """
        return layout

    def do_config(self, config):
        """Update the configuration of the plot.

        Parameters
        ----------
        config : dict
            Plotly config information.

        Returns
        -------
        None

        """
        return config

    def get(self, data):
        """Get the current selection of points.

        Returns
        -------
        list

        """
        return data

    def get_select(self, data):
        """Get the current selection of points.

        Returns
        -------
        list

        """
        return data

    def get_click(self, data):
        """Get the current selection of points.

        Returns
        -------
        list

        """
        return data

    def get_hover(self, data):
        """Get the current selection of points.

        Returns
        -------
        list

        """
        return data

    def get_layout(self, data):
        """Get the current layout.

        Returns
        -------
        list

        """
        return data



"""Test markdown and text widgets."""
# pylint: disable=redefined-outer-name,unused-argument,invalid-name

import time

import pytest
from bowtie import App
from bowtie.html import Markdown
from bowtie.control import Textbox
from bowtie.tests.utils import reset_uuid, server_check


reset_uuid()

mark = Markdown('''
# top
## middle

[link]('hello.html')
''')
side = Markdown('''
# sideheader
''')
text = Textbox(area=True)


def write(txt):
    """Update markdown text."""
    mark.do_text(txt)


@pytest.fixture
def markdown(build_reset, monkeypatch):
    """Create markdown and text widgets."""
    app = App(__name__, sidebar=True)
    app.add(mark)
    app.add_sidebar(side)
    app.add_sidebar(text)
    app.subscribe(text.on_change)(write)
    # pylint: disable=protected-access
    app._build()

    with server_check(app) as server:
        yield server


def test_markdown(markdown, chrome_driver):
    """Test markdown and text widgets."""
    chrome_driver.get('http://localhost:9991')
    chrome_driver.implicitly_wait(5)

    assert chrome_driver.title == 'Bowtie App'

    txtctrl = chrome_driver.find_element_by_class_name('ant-input')
    output = chrome_driver.find_element_by_xpath(
        "//div[@style='grid-area: 1 / 2 / 2 / 3; position: relative;']"
    )

    assert 'top' in output.text
    assert 'middle' in output.text
    assert 'link' in output.text

    txtctrl.send_keys('apple')
    time.sleep(1)

    assert 'apple' in output.text

    txtctrl.send_keys('banana')
    time.sleep(1)

    assert 'apple' in output.text
    assert 'banana' in output.text



"""Authentication out of the box.

References
----------
https://stackoverflow.com/questions/13428708/best-way-to-make-flask-logins-login-required-the-default
https://stackoverflow.com/questions/14367991/flask-before-request-add-exception-for-specific-route

"""

from typing import Dict, Optional
from abc import ABC, abstractmethod

from flask import Response, request, session

from bowtie._app import App


class Auth(ABC):
    """Abstract Authentication class."""

    def __init__(self, app: App) -> None:
        """Create Auth class to protect flask routes and socketio connect."""
        self.app = app
        self.app.app.before_request(self.before_request)
        # only need to check credentials on "connect" event
        self.app._socketio.on('connect')(self.socketio_auth)  # pylint: disable=protected-access

    @abstractmethod
    def before_request(self):
        """Determine if a user is allowed to view this route.

        Name is subject to change.

        Returns
        -------
        None, if no protection is needed.

        """

    @abstractmethod
    def socketio_auth(self) -> bool:
        """Determine if a user is allowed to establish socketio connection.

        Name is subject to change.
        """


class BasicAuth(Auth):
    """Basic Authentication."""

    def __init__(self, app: App, credentials: Dict[str, str]) -> None:
        """
        Create basic auth with credentials.

        Parameters
        ----------
        credentials : dict
            Usernames and passwords should be passed in as a dictionary.

        Examples
        --------
        >>> from bowtie import App
        >>> from bowtie.auth import BasicAuth
        >>> app = App(__name__)
        >>> auth = BasicAuth(app, {'alice': 'secret1', 'bob': 'secret2'})

        """
        self.credentials = credentials
        super().__init__(app)

    def _check_auth(self, username: str, password: str) -> bool:
        """Check if a username/password combination is valid."""
        try:
            return self.credentials[username] == password
        except KeyError:
            return False

    def socketio_auth(self) -> bool:
        """Determine if a user is allowed to establish socketio connection."""
        try:
            return session['logged_in'] in self.credentials
        except KeyError:
            return False

    def before_request(self) -> Optional[Response]:
        """Determine if a user is allowed to view this route."""
        auth = request.authorization
        if not auth or not self._check_auth(auth.username, auth.password):
            return Response(
                'Could not verify your access level for that URL.\n'
                'You have to login with proper credentials', 401,
                {'WWW-Authenticate': 'Basic realm="Login Required"'}
            )
        session['logged_in'] = auth.username
        # pylint wants this return statement
        return None



"""Progress component.

Not for direct use by user.
"""

from bowtie._component import Component


class Progress(Component):
    """Circle progress indicator."""

    _TEMPLATE = 'progress.jsx'
    _COMPONENT = 'AntProgress'
    _PACKAGE = None
    _ATTRS = None

    def __init__(self) -> None:
        """Create a progress indicator.

        This component is used by all visual components.
        It is not meant to be used alone.

        By default, it is not visible.
        It is an opt-in feature and you can happily use Bowtie
        without using the progress indicators at all.

        It is useful for indicating progress to the user for long-running processes.
        It can be accessed through the ``.progress`` accessor.

        Examples
        --------
        Using the progress widget to provide feedback to the user.

        >>> from bowtie.visual import Plotly
        >>> plotly = Plotly()
        >>> def callback(x):
        ...     plotly.progress.do_visible(True)
        ...     plotly.progress.do_percent(0)
        ...     compute1()
        ...     plotly.progress.do_inc(50)
        ...     compute2()
        ...     plotly.progress.do_visible(False)

        References
        ----------
        https://ant.design/components/progress/

        """
        super().__init__()
        self._tagbase = self._tagbase[:-3] + '>'
        self._tags = '<' + self._COMPONENT + self._tagbase, '</AntProgress>'

    @property
    def _instantiate(self) -> str:
        """Instantiate a progress bar.

        This is normally never used.
        """
        return ''.join(self._tags)

    # pylint: disable=no-self-use
    def do_percent(self, percent):
        """Set the percentage of the progress.

        Parameters
        ----------
        percent : number
            Sets the progress to this percentage.

        Returns
        -------
        None

        """
        return percent

    def do_inc(self, inc):
        """Increment the progress indicator.

        Parameters
        ----------
        inc : number
            Value to increment the progress.

        Returns
        -------
        None

        """
        return inc

    def do_visible(self, visible):
        """Hide and shows the progress indicator.

        Parameters
        ----------
        visible : bool
            If ``True`` shows the progress indicator
            otherwise it is hidden.

        Returns
        -------
        None

        """
        return visible

    def do_active(self):
        """Reset the progress to active (in progress) status.

        Returns
        -------
        None

        """

    def do_success(self):
        """Display the progress indicator as done.

        Returns
        -------
        None

        """

    def do_error(self):
        """Display an error in the progress indicator.

        Returns
        -------
        None

        """



"""Bowtie exceptions."""


class YarnError(Exception):
    """Errors from ``Yarn``."""


class WebpackError(Exception):
    """Errors from ``Webpack``."""


class SizeError(Exception):
    """Size values must be a number."""


class GridIndexError(IndexError):
    """Invalid index into the grid layout."""


class NoUnusedCellsError(Exception):
    """All cells are used."""


class SpanOverlapError(Exception):
    """Spans may not overlap."""


class MissingRowOrColumn(Exception):
    """Missing a row or column."""


class NoSidebarError(Exception):
    """Cannot add to the sidebar when it doesn't exist."""


class NotStatefulEvent(Exception):
    """This event is not stateful and cannot be paired with other events."""


class SerializationError(TypeError):
    """Cannot serialize the data for command."""



"""Multiple views testing."""
# pylint: disable=unused-argument,redefined-outer-name,invalid-name

from time import sleep

from numpy import random as rng
import pandas as pd
import pytest

from bowtie import App, View
from bowtie.control import Nouislider, Button
from bowtie.visual import Table
from bowtie.tests.utils import reset_uuid, server_check


reset_uuid()


table = Table()
ctrl = Nouislider()
ctrl2 = Button()


def callback(*args):
    """dummy function"""
    df = pd.DataFrame(rng.randn(10, 10))
    table.do_data(df)


@pytest.fixture
def multiple_views(build_reset, monkeypatch):
    """Create multiple views app."""
    app = App(__name__, sidebar=True)
    view1 = View()  # pylint: disable=unused-variable
    assert view1._uuid == 2  # pylint: disable=protected-access
    view2 = View()
    view2.add(table)
    app.add_route(view2, 'view2')

    app.add(table)
    app.add_sidebar(ctrl)
    app.add_sidebar(ctrl2)
    app.subscribe(ctrl.on_change)(app.subscribe(ctrl2.on_click)(callback))

    app._build()  # pylint: disable=protected-access

    with server_check(app) as server:
        yield server


# pylint: disable=redefined-outer-name,unused-argument
def test_multiple(multiple_views, chrome_driver):
    """Test multiple views app."""
    chrome_driver.get('http://localhost:9991')
    chrome_driver.implicitly_wait(5)

    assert chrome_driver.title == 'Bowtie App'

    button = chrome_driver.find_element_by_class_name('ant-btn')
    data = chrome_driver.find_element_by_class_name('ant-table-body').text
    assert len(data.split('\n')) == 1
    button.click()
    sleep(2)

    data = chrome_driver.find_element_by_class_name('ant-table-body').text
    assert len(data.split('\n')) == 20

    logs = chrome_driver.get_log('browser')
    for log in logs:
        if log['level'] == 'SEVERE':
            raise Exception(log['message'])

    chrome_driver.get('http://localhost:9991/view2')
    data = chrome_driver.find_element_by_class_name('ant-table-body').text

    assert len(data.split('\n')) == 20

    chrome_driver.implicitly_wait(5)

    logs = chrome_driver.get_log('browser')
    for log in logs:
        if log['level'] == 'SEVERE':
            raise Exception(log['message'])

    chrome_driver.get('http://localhost:9991/view1')
    assert chrome_driver.title == '404 Not Found'



"""Interactive Dashboard Toolkit."""

__version__ = '0.11.0'

from bowtie._app import App, View
from bowtie._command import command
from bowtie.pager import Pager
from bowtie._cache import cache


def load_ipython_extension(ipython):
    """Enable IPython extension."""
    from bowtie._magic import BowtieMagic
    ipython.register_magics(BowtieMagic)



"""Test markdown and text widgets."""
# pylint: disable=unused-argument,redefined-outer-name,invalid-name

import time

import pytest
from bowtie import App, cache
from bowtie.control import Button
from bowtie.tests.utils import reset_uuid, server_check


reset_uuid()


button = Button()


def click():
    """Save and load cache."""
    cache['a'] = 3
    assert cache[b'a'] == 3
    assert cache[u'a'] == 3
    cache[b'b'] = True
    assert cache[b'b']
    assert cache[u'b']


def test_keys():
    """Test invalid keys."""
    with pytest.raises(KeyError):
        cache[0] = 0

    with pytest.raises(KeyError):
        cache[True] = 0

    with pytest.raises(KeyError):
        cache[3, 4] = 0


@pytest.fixture
def dummy(build_reset, monkeypatch):
    """Create basic app."""
    app = App(__name__)
    app.add(button)
    app.subscribe(button.on_click)(click)
    app._build()  # pylint: disable=protected-access

    with server_check(app) as server:
        yield server


def test_cache(dummy, chrome_driver):
    """Test cache works."""
    chrome_driver.get('http://localhost:9991')
    chrome_driver.implicitly_wait(5)

    assert chrome_driver.title == 'Bowtie App'

    btx = chrome_driver.find_element_by_class_name('ant-btn')
    btx.click()
    time.sleep(2)



"""Test tag instantation for components."""

from bowtie.html import Markdown


def test_markdown():
    """Test tags for the Markdown widget."""
    # pylint: disable=protected-access
    next_uuid = Markdown._NEXT_UUID
    mark = Markdown()
    assert mark._instantiate == (
        f"<Markdown initial={{''}} socket={{socket}} uuid={{'{next_uuid + 1}'}} />"
    )

    next_uuid = Markdown._NEXT_UUID
    mark = Markdown(initial='#hi\n##hello')
    assert mark._instantiate == (
        "<Markdown initial={'<h1>hi</h1>\\n<h2>hello</h2>'} "
        f"socket={{socket}} uuid={{'{next_uuid + 1}'}} />"
    )



#!/usr/bin/env python
"""Serialization testing."""

import numpy as np
import pandas as pd

from bowtie._component import jdumps, pack


NPARRAY = np.array([5, 6])
NPSCALAR = np.int32(5)
DATES = pd.date_range('2017-01-01', periods=2)


def test_json():
    """Tests json encoding numpy and pandas."""
    assert jdumps(NPARRAY) == jdumps([5, 6])
    assert jdumps(NPSCALAR) == jdumps(5)
    assert jdumps(DATES) == jdumps(['2017-01-01T00:00:00', '2017-01-02T00:00:00'])


def test_msgpack():
    """Tests msgpack encoding numpy and pandas."""
    assert pack(NPARRAY) == pack([5, 6])
    assert pack(NPSCALAR) == pack(5)
    assert pack(DATES) == pack(['2017-01-01T00:00:00', '2017-01-02T00:00:00'])



"""Bowtie cache functions."""

import flask
from flask_socketio import emit
import eventlet
from eventlet.queue import LightQueue
import msgpack

from bowtie._component import pack


def validate(key):
    """Check that the key is a string or bytestring.

    That's the only valid type of key.
    """
    if not isinstance(key, (str, bytes)):
        raise KeyError('Key must be of type str or bytes, found type {}'.format(type(key)))


# pylint: disable=too-few-public-methods
class _Cache:
    """Store data in the browser.

    This cache uses session storage so data will stay
    in the browser until the tab is closed.
    All data must be serializable, which means if the
    serialization transforms the data it won't be the same
    when it is fetched.

    Examples
    --------
    >>> from bowtie import cache
    >>> cache['a'] = True  # doctest: +SKIP
    >>> cache['a']  # doctest: +SKIP
    True
    >>> cache['b'] = np.arange(5)  # doctest: +SKIP
    >>> cache['b']  # doctest: +SKIP
    [1, 2, 3, 4, 5]

    """

    def __getitem__(self, key):
        """Load the value stored with the key.

        Parameters
        ----------
        key : str
            The key to lookup the value stored.

        Returns
        -------
        object
            The value if the key exists in the cache, otherwise None.

        """
        validate(key)
        signal = 'cache_load'
        event = LightQueue(1)
        if flask.has_request_context():
            emit(signal, {'data': pack(key)}, callback=event.put)
        else:
            sio = flask.current_app.extensions['socketio']
            sio.emit(signal, {'data': pack(key)}, callback=event.put)
        return msgpack.unpackb(bytes(event.get(timeout=10)), encoding='utf8')

    def __setitem__(self, key, value):
        """Store the key value pair.

        Parameters
        ----------
        key : str
            The key to determine where it's stored, you'll need this to load the value later.
        value : object
            The value to store in the cache.

        Returns
        -------
        None

        """
        validate(key)
        signal = 'cache_save'
        if flask.has_request_context():
            emit(signal, {'key': pack(key), 'data': pack(value)})
        else:
            sio = flask.current_app.extensions['socketio']
            sio.emit(signal, {'key': pack(key), 'data': pack(value)})
        eventlet.sleep()


# pylint: disable=invalid-name
cache = _Cache()



#!/usr/bin/env python
"""Pytest configuration."""

from sys import platform
import socket
import shutil
from pathlib import Path

from selenium import webdriver
import pytest

from bowtie import View
from bowtie._app import _DIRECTORY


@pytest.fixture
def build_reset():
    """Path for building apps with pytest."""
    sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    result = sock.connect_ex(('127.0.0.1', 9991))
    if result == 0:
        raise Exception('Port 9991 is unavailable, aborting test.')
    # pylint: disable=protected-access
    View._NEXT_UUID = 0
    yield
    shutil.rmtree(Path(__file__).parent / _DIRECTORY)


@pytest.fixture
def chrome_driver():
    """Set up chrome driver."""
    options = webdriver.ChromeOptions()
    if platform == 'darwin':
        options.binary_location = '/Applications/Google Chrome.app/Contents/MacOS/Google Chrome'
    else:
        options.binary_location = '/usr/bin/google-chrome-stable'
    options.add_argument('headless')
    options.add_argument('window-size=1200x800')
    driver = webdriver.Chrome(options=options)
    yield driver
    driver.quit()



"""Test all components for instatiation issues."""
# pylint: disable=unused-argument,redefined-outer-name

from inspect import isclass

import pytest
from bowtie import App
from bowtie import control, visual, html
from bowtie._component import COMPONENT_REGISTRY
from bowtie.tests.utils import reset_uuid, server_check


def create_components():
    """Create components for this test."""
    reset_uuid()
    # pylint: disable=protected-access
    controllers = []
    for compstr in dir(control):
        comp = getattr(control, compstr)
        if (compstr[0] != '_' and isclass(comp) and issubclass(comp, control._Controller)
                and compstr != 'Upload'):
            controllers.append(comp())

    for controller in controllers:
        assert COMPONENT_REGISTRY[controller._uuid] == controller

    visuals = []
    for compstr in dir(visual):
        comp = getattr(visual, compstr)
        if compstr[0] != '_' and isclass(comp) and issubclass(comp, visual._Visual):
            visuals.append(comp())

    for vis in visuals:
        assert COMPONENT_REGISTRY[vis._uuid] == vis

    htmls = []
    for compstr in dir(html):
        comp = getattr(html, compstr)
        if compstr[0] != '_' and isclass(comp) and issubclass(comp, html._HTML):
            htmls.append(comp())

    for htm in htmls:
        assert COMPONENT_REGISTRY[htm._uuid] == htm

    return controllers, visuals, htmls


create_components()


@pytest.fixture
def components(build_reset, monkeypatch):
    """App with all components."""
    controllers, visuals, htmls = create_components()

    app = App(__name__, rows=len(visuals), sidebar=True)
    for controller in controllers:
        # pylint: disable=protected-access
        assert COMPONENT_REGISTRY[controller._uuid] == controller
        app.add_sidebar(controller)

    for vis in visuals:
        # pylint: disable=protected-access
        assert COMPONENT_REGISTRY[vis._uuid] == vis
        app.add(vis)

    for htm in htmls:
        # pylint: disable=protected-access
        assert COMPONENT_REGISTRY[htm._uuid] == htm
        app.add_sidebar(htm)

    assert len(COMPONENT_REGISTRY) == len(controllers) + 2 * len(visuals) + len(htmls)

    # pylint: disable=protected-access
    app._build()

    # run second time to make sure nothing weird happens with subsequent builds
    app._build()

    with server_check(app) as server:
        yield server


def test_components(components, chrome_driver):
    """Test that no components cause an error."""
    chrome_driver.get('http://localhost:9991')
    chrome_driver.implicitly_wait(5)

    logs = chrome_driver.get_log('browser')
    for log in logs:
        if log['level'] == 'SEVERE':
            raise Exception(log['message'])



"""
Decorates a function for Bowtie.

Reference
---------
https://gist.github.com/carlsmith/800cbe3e11f630ac8aa0
"""

from typing import Callable
import sys
import inspect
from subprocess import call

import click

from bowtie._app import _WEBPACK, App


class WrongNumberOfArguments(TypeError):
    """The "build" function accepts an incorrect number of arguments."""


def numargs(func: Callable) -> int:
    """Get number of arguments."""
    return len(inspect.signature(func).parameters)


def command(func):
    """Command line interface decorator.

    Decorate a function for building a Bowtie
    application and turn it into a command line interface.
    """
    # pylint: disable=protected-access,unused-variable
    nargs = numargs(func)
    if nargs > 0:
        raise WrongNumberOfArguments(
            f'Decorated function "{func.__name__}" should have no arguments, it has {nargs}.'
        )
    app = func()
    if app is None:
        raise TypeError(
            'No `App` instance was returned. '
            'In the function decorated with @command, '
            'return the `App` instance so it can be built.'
        )
    if not isinstance(app, App):
        raise TypeError(
            f'Returned value {app} is of type {type(app)}, '
            'it needs to be a bowtie.App instance.'
        )

    @click.group(options_metavar='[--help]')
    def cmd():
        """Bowtie CLI to help build and run your app."""
        pass

    @cmd.command(add_help_option=False)
    def build():
        """Write the app, downloads the packages, and bundles it with Webpack."""
        app._build()

    @cmd.command(add_help_option=False)
    @click.option('--host', '-h', default='0.0.0.0', type=str)
    @click.option('--port', '-p', default=9991, type=int)
    def run(host, port):
        """Build the app and serve it."""
        app._build()
        app._serve(host, port)

    @cmd.command(add_help_option=True)
    @click.option('--host', '-h', default='0.0.0.0', type=str)
    @click.option('--port', '-p', default=9991, type=int)
    def serve(host, port):
        """Serve the Bowtie app."""
        app._serve(host, port)

    @cmd.command(context_settings=dict(ignore_unknown_options=True),
                 add_help_option=False)
    @click.argument('extra', nargs=-1, type=click.UNPROCESSED)
    def dev(extra):
        """Recompile the app for development."""
        line = (_WEBPACK, '--config', 'webpack.dev.js') + extra
        call(line, cwd=app._build_dir)

    @cmd.command(context_settings=dict(ignore_unknown_options=True),
                 add_help_option=False)
    @click.argument('extra', nargs=-1, type=click.UNPROCESSED)
    def prod(extra):
        """Recompile the app for production."""
        line = (_WEBPACK, '--config', 'webpack.prod.js', '--progress') + extra
        call(line, cwd=app._build_dir)

    locale = inspect.stack()[1][0].f_locals
    module = locale.get("__name__")

    if module == "__main__":
        try:
            arg = sys.argv[1:]
        except IndexError:
            arg = ('--help',)
        # pylint: disable=too-many-function-args
        sys.exit(cmd(arg))

    return cmd



"""App and View testing."""
import pytest

from bowtie import App


def test_subscribe_error():
    """Subscribe with incorrect argument order."""
    app = App()
    with pytest.raises(IndexError):
        app.subscribe()



"""Static HTML Components."""

from flask import Markup
from markdown import markdown

from bowtie._component import Component


# pylint: disable=too-few-public-methods
class _HTML(Component):
    """Abstract class for HTML components."""

    # pylint: disable=abstract-method
    @property
    def _instantiate(self) -> str:
        tagwrap = '{component}' + self._tagbase
        return self._insert(tagwrap, self._comp)


# pylint: disable=too-few-public-methods
class Markdown(_HTML):
    """Display Markdown."""

    _TEMPLATE = 'markdown.jsx'
    _COMPONENT = 'Markdown'
    _PACKAGE = None
    _ATTRS = "initial={{'{initial}'}}"

    def __init__(self, initial: str = '') -> None:
        """Create a Markdown widget.

        Parameters
        ----------
        initial : str, optional
            Default markdown for the widget.

        """
        super().__init__()
        self._comp = self._tag.format(
            initial=Markup(markdown(initial).replace('\n', '\\n'))
        )

    # pylint: disable=no-self-use
    def do_text(self, text):
        """Replace widget with this text.

        Parameters
        ----------
        test : str
            Markdown text as a string.

        Returns
        -------
        None

        """
        return Markup(markdown(text))

    def get(self, text):
        """Get the current text.

        Returns
        -------
        String of html.

        """
        return text


class Link(_HTML):
    """An internal link.

    This doesn't create a page reload.
    """

    _TEMPLATE = 'link.jsx'
    _COMPONENT = 'ALink'
    _PACKAGE = None
    _ATTRS = "to={{'{link}'}}"

    def __init__(self, link: str = '/') -> None:
        """Create a button.

        Parameters
        ----------
        link : str

        """
        super().__init__()
        self._comp = self._tag.format(
            link=link
        )


class Div(_HTML):
    """Div tag."""

    _TEMPLATE = 'div.jsx'
    _COMPONENT = 'Bowdiv'
    _PACKAGE = None
    _ATTRS = "initial={{'{initial}'}}"

    def __init__(self, text: str = '') -> None:
        """Create header text with a size.

        Parameters
        ----------
        text : str
            Text of the header tag.

        """
        super().__init__()
        self._comp = self._tag.format(
            initial=text
        )

    # pylint: disable=no-self-use
    def do_text(self, text):
        """Replace widget with this text.

        Parameters
        ----------
        test : str
            Markdown text as a string.

        Returns
        -------
        None

        """
        return text

    def get(self, text):
        """Get the current text.

        Returns
        -------
        String of html.

        """
        return text


class Header(_HTML):
    """Header tag."""

    _TEMPLATE = 'header.jsx'
    _COMPONENT = 'Bowhead'
    _PACKAGE = None
    _ATTRS = ("initial={{'{initial}'}} "
              'size={{{size}}}')

    def __init__(self, text: str = '', size: int = 1) -> None:
        """Create header text with a size.

        Parameters
        ----------
        text : str
            Text of the header tag.
        size : int
            Size of the header tag from 1 to 6.

        """
        super().__init__()
        if size not in range(1, 7):
            raise ValueError('Header size must be in [1..6], found {}.'.format(size))
        self._comp = self._tag.format(
            initial=text,
            size=size
        )

    # pylint: disable=no-self-use
    def do_text(self, text):
        """Replace widget with this text.

        Parameters
        ----------
        test : str
            Markdown text as a string.

        Returns
        -------
        None

        """
        return text

    def get(self, text):
        """Get the current text.

        Returns
        -------
        String of html.

        """
        return text
